import torch
from GANDLF.grad_clipping.clip_gradients import dispatch_clip_grad_


class GradScaler:
    def __init__(self):
        """Initializes a GradScaler object with a PyTorch GradScaler."""
        self._scaler = torch.cuda.amp.GradScaler()

    def __call__(
        self,
        loss,
        optimizer,
        clip_grad=None,
        clip_mode="norm",
        parameters=None,
        create_graph=False,
    ):
        """
        Scales the loss and performs backward pass through the computation graph.

        Args:
            loss (torch.Tensor): The loss tensor to scale and backpropagate.
            optimizer (torch.optim.Optimizer): The optimizer to step after backpropagation.
            clip_grad (float): The clipping value/factor/norm, mode dependent (default: None).
            clip_mode (str): The clipping mode, one of 'norm', 'value', 'agc' (default: 'norm').
            parameters (Iterable): The model parameters to clip (default: None).
            create_graph (bool): Whether to create a new graph for backpropagation (default: False).
        """
        self._scaler.scale(loss).backward(create_graph=create_graph)
        if clip_grad is not None:
            assert parameters is not None
            # unscale the gradients of optimizer's assigned params in-place
            self._scaler.unscale_(optimizer)
            if (clip_mode is None) or (str(clip_mode).lower() == "none"):
                clip_mode = "norm"  # default, in case none gets passed
            dispatch_clip_grad_(parameters, clip_grad, mode=clip_mode)
        self._scaler.step(optimizer)
        self._scaler.update()

    def state_dict(self):
        """
        Returns the state dict of the underlying GradScaler.
        """
        return self._scaler.state_dict()

    def load_state_dict(self, state_dict):
        """
        Loads the state dict of the underlying GradScaler.

        Args:
            state_dict (dict): The state dict to load.
        """
        self._scaler.load_state_dict(state_dict)


def model_parameters_exclude_head(model, clip_mode=None):
    """
    Returns the parameters of a PyTorch model excluding the last two layers (the head).

    Args:
        model (torch.nn.Module): The PyTorch model to get the parameters from.
        clip_mode (str): The clipping mode, one of 'norm', 'value', 'agc' (default: None).

    Returns:
        Iterable: The model parameters excluding the last two layers if clip_mode is 'agc', otherwise all parameters.
    """
    exclude_head = False
    if clip_mode is not None:
        if clip_mode == "agc":
            exclude_head = True
    if exclude_head:
        return [p for p in model.parameters()][:-2]
    else:
        return model.parameters()



import os

# hides torchio citation request, see https://github.com/fepegar/torchio/issues/235
os.environ["TORCHIO_HIDE_CITATION_PROMPT"] = "1"

from .imaging import (
    resize_image,
    resample_image,
    perform_sanity_check_on_subject,
    write_training_patches,
)

from .tensor import (
    one_hot,
    reverse_one_hot,
    send_model_to_device,
    get_model_dict,
    get_class_imbalance_weights,
    get_class_imbalance_weights_segmentation,
    get_class_imbalance_weights_classification,
    get_linear_interpolation_mode,
    print_model_summary,
    get_ground_truths_and_predictions_tensor,
    get_output_from_calculator,
    get_tensor_from_image,
    get_image_from_tensor,
)

from .write_parse import (
    writeTrainingCSV,
    parseTrainingCSV,
    parseTestingCSV,
    get_dataframe,
    convert_relative_paths_in_dataframe,
)

from .parameter_processing import (
    populate_header_in_parameters,
    find_problem_type,
    find_problem_type_from_parameters,
    populate_channel_keys_in_params,
)

from .generic import (
    get_date_time,
    get_unique_timestamp,
    get_filename_extension_sanitized,
    version_check,
    get_array_from_image_or_tensor,
    suppress_stdout_stderr,
    set_determinism,
)

from .modelio import (
    best_model_path_end,
    latest_model_path_end,
    initial_model_path_end,
    load_model,
    load_ov_model,
    save_model,
    optimize_and_save_model,
)
